import logging


class DuplicationHandler:
    instance = None

    def __init__(self, mode: str):
        self.mode = mode.strip().lower()
        self.names = list()

    def should_generate(self, name: str):
        if self.mode == "once":
            if name in self.names:
                logging.warning(f"Skipping {name}, already written earlier")
                return False
        return True

    def store_generated(self, name: str):
        self.names.append(name)


def get() -> DuplicationHandler:
    return DuplicationHandler.instance


def init(mode: str):
    DuplicationHandler.instance = DuplicationHandler(mode)


def should_generate(name) -> bool:
    return get().should_generate(name)


def store_generated(name):
    return get().store_generated(name)



from typing import Any

from prance import ResolvingParser
from prance.util.resolver import RESOLVE_HTTP, RESOLVE_FILES, TRANSLATE_EXTERNAL

from caffoa.function_writer import FunctionWriter
from caffoa.interface_writer import InterfaceWriter
from caffoa.model_parser import ModelParser
from caffoa.model_writer import ModelWriter
from caffoa.path_parser import PathParser


class OpenApiFile:
    def __init__(self, name: str, version: int, config: dict):
        self.version = version
        self.name = name
        self.base_config = config
        self._model_parser = None
        self._function_parser = None
        self.model = None
        self.imports = list()
        self.known_types = dict()

    def get_config(self, config: dict, name: str, default_value: Any = None):
        return config.get(name, self.base_config.get(name, default_value))

    def model_parser(self):
        if self._model_parser is None:
            self._model_parser = ResolvingParser(self.name, strict=False, resolve_types=RESOLVE_HTTP | RESOLVE_FILES,
                                                 resolve_method=TRANSLATE_EXTERNAL)
            try:
                schemas = self._model_parser.specification["components"]["schemas"]
                for class_name in list(schemas.keys()):
                    if ".yml_schemas_" in class_name:
                        new_class_name = class_name.split(".yml_schemas_")[-1]
                        schemas[new_class_name] = schemas[class_name]
                        del schemas[class_name]
            except KeyError:
                pass  # no schema

        return self._model_parser

    def function_parser(self) -> ResolvingParser:
        if self._function_parser is None:
            self._function_parser = ResolvingParser(self.name, strict=False)
        return self._function_parser

    def create_model(self, config: dict):
        if not "namespace" in config or not "targetFolder" in config:
            raise Warning(f"model needs children 'namespace' and 'targetFolder' in service #{id}")
        parser = ModelParser()
        parser.prefix = config.get("prefix", self.base_config.get("prefix", ""))
        parser.suffix = config.get("suffix", self.base_config.get("suffix", ""))

        parser.excludes = list(config.get('excludes', list()))
        parser.includes = list(config.get('includes', list()))
        model = parser.parse(self.model_parser())
        writer = ModelWriter(self.version, config["namespace"], config["targetFolder"])
        writer.additional_imports.extend(config.get("imports", self.base_config.get('imports', list())))
        writer.enum_name = self.get_config(config, "enumName", writer.enum_name)
        writer.enum_list_name = self.get_config(config, "enumListName", writer.enum_list_name)
        writer.check_enums = self.get_config(config, "checkEnums", writer.check_enums)

        writer.write(model)
        self.imports.append(writer.namespace)
        self.known_types = parser.known_types

    def parse_endpoints(self, create_returns: bool):
        parser = PathParser(self.function_parser())
        parser.known_types = self.known_types
        parser.prefix = self.base_config.get("prefix", "")
        parser.suffix = self.base_config.get("suffix", "")
        if create_returns:
            parser.create_returns(self.model_parser())
            parser.create_bodies(self.model_parser())
        return parser.parse()

    def create_function(self, config: dict):
        if not "name" in config or not "namespace" in config or not "targetFolder" in config:
            raise Warning(f"function needs children 'name', 'namespace' and 'targetFolder' in service {self.name}")

        endpoints = self.parse_endpoints(self.version > 2)

        name = config['name']
        namespace = config["namespace"]
        target_folder = config['targetFolder']

        iwriter = InterfaceWriter(self.version, name, namespace, target_folder)
        iwriter.use_factory = self.get_config(config, "useFactory", iwriter.use_factory)

        iwriter.imports.extend(self.get_config(config, "imports", list()))
        iwriter.imports.extend(self.imports)
        iwriter.interface_name = self.get_config(config, 'interfaceName', iwriter.interface_name)
        iwriter.namespace = self.get_config(config, 'interfaceNamespace', iwriter.namespace)
        iwriter.target_folder = self.get_config(config, 'interfaceTargetFolder', iwriter.target_folder)
        iwriter.request_body_filter = self.get_config(config, 'requestBodyType', iwriter.request_body_filter)
        iwriter.write(endpoints)

        writer = FunctionWriter(self.version, name, namespace, target_folder, iwriter.interface_name)
        writer.use_factory = self.get_config(config, 'useFactory', writer.use_factory)
        writer.boilerplate = self.get_config(config, 'boilerplate', writer.boilerplate)
        writer.functions_name = self.get_config(config, 'functionsName', writer.functions_name)
        writer.error_namespace = self.get_config(config, "errorNamespace", writer.error_namespace)
        writer.json_error_handling = self.get_config(config, "jsonErrorHandling", writer.json_error_handling)
        writer.error_folder = self.get_config(config, "errorFolder", writer.error_folder)
        writer.imports.extend(self.get_config(config, "imports", list()))
        writer.request_body_filter = self.get_config(config, 'requestBodyType', list())
        if self.version > 2:
            writer.imports.extend(self.imports)
        writer.write(endpoints)



import logging
from typing import Tuple, Optional, List, Set
import re

from prance import ResolvingParser

from caffoa.converter import parse_type, to_camelcase, is_primitive
from caffoa.model import ModelData, MemberData
from caffoa.object_parser import ObjectParser


class ModelParser:
    def __init__(self):
        self.prefix = ""
        self.suffix = ""
        self.known_types = dict()
        self.excludes = list()
        self.includes = list()

    def parse(self, parser: ResolvingParser) -> List[ModelData]:
        schemas = parser.specification["components"]["schemas"]
        self.parse_simple_types(schemas)
        objects = self.parse_objects(schemas)
        return self.add_interfaces(objects)

    def parse_simple_types(self, schemas: dict):
        for class_name, schema in schemas.items():
            if class_name in self.excludes:
                continue
            if len(self.includes) > 0 and class_name not in self.includes:
                continue
            if "type" in schema and is_primitive(schema["type"]):
                class_name = self.class_name(class_name)
                self.known_types[class_name] = schema

    def parse_objects(self, schemas: dict) -> List[ModelData]:
        objects = list()
        for name, schema in schemas.items():
            if self.class_name(name) in self.known_types:
                continue
            if name in self.excludes:
                continue
            if len(self.includes) > 0 and name not in self.includes:
                continue
            objects.append(ObjectParser(name, self.prefix, self.suffix, self.known_types).parse(schema))
        return objects

    def class_name(self, name):
        return self.prefix + to_camelcase(name) + self.suffix

    def add_interfaces(self, objects: List[ModelData]) -> list:
        interfaces = dict()
        for item in objects:
            if item.is_interface():
                for child in item.children:
                    if child not in interfaces:
                        interfaces[child] = list()
                    interfaces[child].append(item.name)
        for item in objects:
            if item.name in interfaces:
                item.interfaces = interfaces[item.name]
        return objects



from typing import Optional

from caffoa.converter import is_primitive, parse_type
from caffoa.model import BodyConfig
from caffoa.object_parser import BaseObjectParser


class ContentParser:
    def __init__(self, known_types: dict, prefix: str, suffix: str):
        self.known_types = known_types
        self.suffix = suffix
        self.prefix = prefix

    def parse_content_list(self, content: dict, one_of_schemas: dict):
        schema = self.get_schema(content)
        if schema is None:
            return None
        if "$ref" in schema:
            typename = schema["$ref"].split('/')[-1]
            if typename in one_of_schemas:
                schema = one_of_schemas[typename]

        if "oneOf" in schema:
            if "discriminator" not in schema:
                raise Warning("Need discriminator in oneOf")
            discriminator = schema["discriminator"]
            pre_mapping = dict()
            for e_value, e_type in discriminator.get("mapping", dict()).items():
                pre_mapping[e_type] = e_value
            names = list()
            mapping = dict()
            for element in schema["oneOf"]:
                if "$ref" not in element:
                    raise Warning("Cannot have oneOf without ref types")
                ref = element["$ref"]
                name = self.name_for_ref(ref)
                names.append(name)
                if ref in pre_mapping:
                    map_name = pre_mapping[ref]
                else:
                    map_name = ref.split('/')[-1]
                mapping[map_name] = name
            config = BodyConfig(names)
            config.mapping = mapping
            config.discriminator = discriminator["propertyName"]
            return config

        return BodyConfig([self.parse_element(schema)])

    def parse_content_string(self, content: dict):
        schema = self.get_schema(content)
        if schema is None:
            return None
        return self.parse_element(schema)

    def parse_element(self, schema: dict):
        if "$ref" in schema:
            return self.name_for_ref(schema["$ref"])
        if "type" in schema and is_primitive(schema["type"]):
            return parse_type(schema)
        if "type" in schema and schema["type"].lower() == "array":
            if "$ref" in schema["items"]:
                name = self.name_for_ref(schema["items"]["$ref"])
                return f"IEnumerable<{name}>"
            if "type" in schema["items"] and is_primitive(schema["items"]["type"]):
                type = parse_type(schema["items"])
                return f"IEnumerable<{type}>"
            raise Warning("array with complex type that is not a reference")

        raise Warning(
            "complex type. Only array, ref or basic types are supported.")

    @staticmethod
    def get_schema(content: dict) -> Optional[dict]:
        if "content" in content:
            content = content["content"]
            keys = list(content.keys())
            if len(keys) > 1:
                raise Warning(
                    f"multiple possible responses")
            type = str(keys[0])
            if type.lower() != "application/json":
                raise Warning(
                    f"type {type}. Only application/json is currently supported for")
            try:
                schema = content[type]['schema']
                return schema
            except KeyError:
                raise Warning(
                    f"type {type}. no schema found")
        return None

    def name_for_ref(self, ref):
        name = BaseObjectParser(self.prefix, self.suffix).class_name_from_ref(ref)
        if name in self.known_types:
            return dict(self.known_types[name])["type"]
        return name



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='caffoa',
    version='0.31.0',
    packages=['caffoa'],
    package_data={
        'caffoa': ['./data/**/*'],
    },
    include_package_data=True,
    url='https://github.com/claasd/caffoa',
    license='MIT',
    author='Claas Diederichs',
    author_email='',
    description='Create Azure Functions From Open Api (for C#)',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        "prance",
        "openapi-spec-validator",
        "PyYaml"
    ]
)



import logging
import os
from typing import List

from caffoa import duplication_handler
from caffoa.base_writer import BaseWriter
from caffoa.converter import to_camelcase
from caffoa.model import ModelData, MemberData, ModelObjectData, ModelInterfaceData


class ModelWriter(BaseWriter):
    def __init__(self, version: int, namespace: str, output_folder: str):
        super().__init__(version)
        self.output_folder = output_folder
        self.namespace = namespace
        self.additional_imports = list()

        self.model_template = self.load_template("ModelTemplate.cs")
        self.prop_template = self.load_template("ModelPropertyTemplate.cs")
        self.enum_prop_template = self.load_template("ModelEnumPropertyTemplate.cs")
        self.interface_template = self.load_template("ModelInterfaceTemplate.cs")
        self.enum_name = "{uFieldName}{uValueName}Value"
        self.enum_list_name = "AllowedValuesFor{uFieldName}"
        self.check_enums = True if version > 2 else False

    def write(self, models: List[ModelData]):
        os.makedirs(self.output_folder, exist_ok=True)
        dates_in_models = False
        for model in models:
            if duplication_handler.should_generate(model.name):
                if model.is_interface():
                    self.write_interface(model)
                else:
                    self.write_model(model)
                    if self.dates_in_model(model):
                        dates_in_models = True
                duplication_handler.store_generated(model.name)

        if dates_in_models:
            self.write_custom_date_converter()

    @staticmethod
    def dates_in_model(model: ModelData):
        for prop in model.properties:
            if prop.is_date:
                return True
        return False

    def write_model(self, model: ModelObjectData):
        imports = [f"using {key};\n" for key in model.imports]
        imports.extend([f"using {key};\n" for key in self.additional_imports])
        # remove duplicates but keep order:
        imports = list(dict.fromkeys(imports))
        all_parents = model.interfaces
        if model.parent:
            all_parents.insert(0, model.parent)
        parent = f" : {', '.join(all_parents)}" if all_parents else ""
        formatted_properties = []
        for prop in model.properties:
            formatted_properties.append(self.format_property(prop))
        formatted_updater = self.format_updater(model)

        file_name = os.path.abspath(self.output_folder + f"/{model.name}.generated.cs")
        logging.info(f"Writing class {model.name} -> {file_name}")

        description = "/// AUTOGENERED BY caffoa ///\n\t"
        if model.description is not None:
            model_description = model.description.strip().replace("\n", "\n\t/// ")
            description += f"/// <summary>\n\t/// {model_description}\n\t/// </summary>\n\t"

        with open(file_name, "w", encoding="utf-8") as f:
            f.write(self.model_template.format(NAMESPACE=self.namespace,
                                               NAME=model.name,
                                               RAWNAME=model.rawname,
                                               PROPERTIES="\n\n".join(formatted_properties),
                                               UPDATEPROPS=formatted_updater,
                                               IMPORTS="".join(imports),
                                               PARENTS=parent,
                                               DESCRIPTION=description))

    def write_interface(self, model: ModelInterfaceData):
        file_name = os.path.abspath(self.output_folder + f"/{model.name}.generated.cs")
        logging.info(f"Writing class {model.name} -> {file_name}")
        description = "/// AUTOGENERED BY caffoa ///\n\t"
        if model.description is not None:
            model_description = model.description.strip().replace("\n", "\n\t/// ")
            description += f"/// <summary>\n\t/// {model_description}\n\t/// </summary>\n\t"
        with open(file_name, "w", encoding="utf-8") as f:
            f.write(self.interface_template.format(
                NAMESPACE=self.namespace,
                NAME=model.name,
                DESCRIPTION=description,
                TYPE=model.discriminator))

    def format_property(self, property: MemberData) -> str:
        extra = ""
        if property.is_date:
            extra = "[JsonConverter(typeof(CustomJsonDateConverter))]\n\t\t"
        default_str = ""
        if property.default_value is not None:
            default_str = f" = {property.default_value};"

        json_property_extra = ""
        if property.is_required and property.nullable:
            json_property_extra = ", Required = Required.AllowNull"
        elif property.is_required:
            json_property_extra = ", Required = Required.Always"

        description = ""
        if property.description is not None:
            description = property.description.strip().replace("\n", "\n\t\t/// ")
            description = f"/// <summary>\n\t\t/// {description}\n\t\t/// </summary>\n\t\t"

        template_data = dict(TYPE=property.typename,
                             NAMELOWER=property.name,
                             NAMEUPPER=to_camelcase(property.name),
                             JSON_EXTRA=extra,
                             DEFAULT=default_str,
                             JSON_PROPERTY_EXTRA=json_property_extra,
                             DESCRIPTION=description)

        if property.enums is not None:
            formatted_enums = {value: self.format_enum_value(property.name, name)
                               for value, name in property.enums.items() if value is not None}
            enums = [f"public const {property.typename.rstrip('?')} {name} = {value};" for value, name in formatted_enums.items()]
            enums = f"\n\t\t".join(enums)
            all_enums = ", ".join(formatted_enums.values())
            if None in property.enums.keys():
                all_enums += ", null"
            if property.default_value is not None:
                template_data["DEFAULT"] = " = " + formatted_enums.get(property.default_value, property.default_value)
            enum_list_name = self.enum_list_name.format(uFieldName=to_camelcase(property.name),
                                                        lFieldName=property.name)
            template = self.enum_prop_template
            template_data['ENUMS'] = enums
            template_data['ENUM_NAMES'] = all_enums
            template_data['ENUM_LIST_NAME'] = enum_list_name
            if self.check_enums:
                template_data['ENUM_CHECK'] = f'if(!{enum_list_name}.Contains(value))\n\t\t\t\t\tthrow new ArgumentOutOfRangeException("{to_camelcase(property.name)}", $"{{value}} is not allowed. Allowed values: {{{enum_list_name}}}");'
            else:
                template_data['ENUM_CHECK'] = "// set checkEnums=true in config file to have a value check here"

        else:
            template = self.prop_template

        return template.format_map(template_data)

    def format_updater(self, model: ModelData) -> str:
        props_update = []
        for prop in model.properties:
            name = to_camelcase(prop.name)
            special_call = ""
            if prop.is_list:
                if prop.is_basic_type:
                    special_call = ".ToList()"
            elif not prop.is_basic_type:
                special_call = f".To{prop.typename}()"
                if prop.nullable:
                    special_call = "?" + special_call
            props_update.append(f"{name} = other.{name}{special_call};")
        formatted = "\n\t\t\t".join(props_update)
        if model.parent:
            formatted = f"UpdateWith{model.parent}(other);\n\t\t\t{formatted}"
        return formatted

    def write_custom_date_converter(self):
        template = self.load_template("DateSerializer.cs")
        file_name = os.path.abspath(self.output_folder + f"/CustomJsonDateConverter.generated.cs")
        logging.info(f"Writing CustomJsonDateConverter -> {file_name}")
        with open(file_name, "w", encoding="utf-8") as f:
            f.write(template.format(NAMESPACE=self.namespace))

    def format_enum_value(self, typename: str, valuename: str):
        return self.enum_name.format(
            lFieldName=typename,
            uFieldName=to_camelcase(typename),
            lValueName=valuename,
            uValueName=valuename if valuename[0].isnumeric() else to_camelcase(valuename)
        )



import logging
import os
from typing import List, Optional

from caffoa.model import Response, MethodResult, EndPoint


def capitalize_first(data: str) -> str:
    if len(data) > 1:
        return data[0].upper() + data[1:]
    return data


def to_camelcase(s):
    s = s.replace("-", "_")
    return ''.join(capitalize_first(x) or '_' for x in s.split('_'))


def is_date(schema: dict):
    if "format" not in schema or "type" not in schema:
        return False
    return schema['type'] == "string" and schema['format'] == "date"


def is_primitive(type: str):
    return type.lower() in ["string", "integer", "number", "boolean"]


def parse_type(schema: dict, nullable: bool = False) -> str:
    format = None
    if "format" in schema:
        format = schema['format'].lower()
    type = schema['type'].lower()
    if nullable:
        suffix = "?"
    else:
        suffix = ""
    if type == "string" and format == "uuid":
        return f"System.Guid{suffix}"
    if type == "string" and format in ["date-time", "date"]:
        return f"System.DateTime{suffix}"
    if type == "integer" and format == "int64":
        return f"long{suffix}"
    if type == "integer" and format == "uint64":
        return f"ulong{suffix}"
    if type == "integer" and format == "uint32":
        return f"uint{suffix}"
    if type == "integer":
        return f"int{suffix}"
    if type == "number":
        return f"double{suffix}"
    if type == "boolean":
        return f"bool{suffix}"
    if type == "string":
        return "string"  # no nullable suffix for string
    return f"{type}{suffix}"


def get_response_type(endpoint: EndPoint) -> Optional[MethodResult]:
    responses = endpoint.responses
    type = None
    codes = list()
    if responses is None:
        return None
    for response in responses:
        if 200 <= response.code < 300:
            if type is not None and type != response.content:
                logging.warning("Returning different objects is not supported")
                return None
            type = response.content
            codes.append(response.code)
    if len(codes) == 0:
        return None
    result = MethodResult(type)
    if len(codes) == 1:
        result = MethodResult(type)
        result.code = codes[0]
        return result
    if type is None:
        result.name = "int"
    else:
        result.name = f'({type}, int)'
    result.is_simple = False
    result.codes = codes
    return result
